#!/usr/bin/env python
"""
This is the unittest for pack module.

python -m unittest -v tests/test_pack.py
python -m pytest --cov=pyjams --cov-report term-missing -v tests/test_pack.py

265: Index creation failing is probably not possible
     at that stage of the algorithm

"""
import unittest


class TestPack(unittest.TestCase):
    """
    Tests for pack.py
    """

    def test_pack(self):
        import numpy as np
        from pyjams import pack

        # dimensions
        d0 = 2
        d1 = 3
        d2 = 4
        d3 = 5

        # data arrays
        s2 = (d1, d0)
        data2d = np.arange(np.prod(s2), dtype=float).reshape(s2) + 1.
        s3 = (d2, d1, d0)
        data3d = np.arange(np.prod(s3), dtype=np.float32).reshape(s3) + 1.
        s4 = (d3, d2, d1, d0)
        data4d = np.arange(np.prod(s4), dtype=int).reshape(s4) + 1

        # masked arrays
        mask2d = np.zeros(np.prod(s2), dtype=bool)
        mask2d[::2] = True
        mask2d = mask2d.reshape(s2)

        mask3d = np.zeros(np.prod(s3), dtype=bool)
        mask3d[::2] = True
        mask3d = mask3d.reshape(s3)

        # masked array as values
        imask2d = mask2d.astype(int)
        fmask2d = mask2d.astype(float)

        # data2d, mask2d
        ois = pack(data2d, mask2d)
        ii = np.where(mask2d)
        osoll = data2d[ii]
        assert ois.dtype == data2d.dtype
        assert np.all(ois == osoll)

        # data3d, mask3d
        ois = pack(data3d, mask3d)
        ii = np.where(mask3d)
        osoll = data3d[ii]
        assert ois.dtype == data3d.dtype
        assert np.all(ois == osoll)

        # data3d, mask2d
        ois = pack(data3d, mask2d)
        ii = np.where(mask2d)
        osoll = data3d[:, ii[0], ii[1]]
        assert ois.dtype == data3d.dtype
        assert np.all(ois == osoll)

        # data3d, mask2d as values
        ois = pack(data3d, imask2d)
        ii = np.where(mask2d)
        osoll = data3d[:, ii[0], ii[1]]
        assert ois.dtype == data3d.dtype
        assert np.all(ois == osoll)
        ois = pack(data3d, fmask2d)
        assert ois.dtype == data3d.dtype
        assert np.all(ois == osoll)

        # data4d, mask2d
        ois = pack(data4d, mask2d)
        ii = np.where(mask2d)
        osoll = data4d[..., ii[0], ii[1]]
        assert ois.dtype == data4d.dtype
        assert np.all(ois == osoll)

        # data4d, mask3d
        ois = pack(data4d, mask3d)
        ii = np.where(mask3d)
        osoll = data4d[:, ii[0], ii[1], ii[2]]
        assert ois.dtype == data4d.dtype
        assert np.all(ois == osoll)

    def test_unpack(self):
        import numpy as np
        from pyjams import pack, unpack

        # dimensions
        d0 = 2
        d1 = 3
        d2 = 4
        d3 = 5

        # data arrays
        s2 = (d1, d0)
        data2d = np.arange(np.prod(s2)).reshape(s2) + 1.
        s3 = (d2, d1, d0)
        data3d = np.arange(np.prod(s3)).reshape(s3) + 1.
        s4 = (d3, d2, d1, d0)
        data4d = np.arange(np.prod(s4), dtype=int).reshape(s4) + 1

        # masked arrays
        mask2d = np.zeros(np.prod(s2), dtype=bool)
        mask2d[::2] = True
        mask2d = mask2d.reshape(s2)

        mask3d = np.zeros(np.prod(s3), dtype=bool)
        mask3d[::2] = True
        mask3d = mask3d.reshape(s3)

        # masked array as values
        imask2d = mask2d.astype(int)
        fmask2d = mask2d.astype(float)

        # data2d, mask2d
        ois = pack(data2d, mask2d)
        ii = np.where(mask2d)
        osoll = data2d[ii]
        assert ois.dtype == data2d.dtype
        assert np.all(ois == osoll)
        uis = unpack(ois, mask2d)
        assert uis.dtype == data2d.dtype
        assert np.all(uis[ii] == data2d[ii])
        uis = unpack(ois, mask2d, fill_value=-1)
        jj = np.where(~mask2d)
        assert uis.dtype == data2d.dtype
        assert np.all(uis[ii] == data2d[ii])
        assert np.all(uis[jj] == -1)

        # data3d, mask3d
        ois = pack(data3d, mask3d)
        ii = np.where(mask3d)
        osoll = data3d[ii]
        assert ois.dtype == data3d.dtype
        assert np.all(ois == osoll)
        uis = unpack(ois, mask3d)
        assert uis.dtype == data3d.dtype
        assert np.all(uis[ii] == data3d[ii])
        uis = unpack(ois, mask3d, -1)
        jj = np.where(~mask3d)
        assert uis.dtype == data3d.dtype
        assert np.all(uis[ii] == data3d[ii])
        assert np.all(uis[jj] == -1)

        # data3d, mask2d
        ois = pack(data3d, mask2d)
        ii = np.where(mask2d)
        osoll = data3d[:, ii[0], ii[1]]
        assert ois.dtype == data3d.dtype
        assert np.all(ois == osoll)
        uis = unpack(ois, mask2d)
        assert uis.dtype == data3d.dtype
        assert np.all(uis[:, ii[0], ii[1]] == data3d[:, ii[0], ii[1]])
        uis = unpack(ois, mask2d, -1)
        jj = np.where(~mask2d)
        assert uis.dtype == data3d.dtype
        assert np.all(uis[:, ii[0], ii[1]] == data3d[:, ii[0], ii[1]])
        assert np.all(uis[:, jj[0], jj[1]] == -1)

        # data3d, mask2d as values
        ois = pack(data3d, imask2d)
        ii = np.where(mask2d)
        osoll = data3d[:, ii[0], ii[1]]
        assert ois.dtype == data3d.dtype
        assert np.all(ois == osoll)
        uis = unpack(ois, imask2d)
        assert uis.dtype == data3d.dtype
        assert np.all(uis[:, ii[0], ii[1]] == data3d[:, ii[0], ii[1]])
        uis = unpack(ois, imask2d, -1)
        jj = np.where(~mask2d)
        assert uis.dtype == data3d.dtype
        assert np.all(uis[:, ii[0], ii[1]] == data3d[:, ii[0], ii[1]])
        assert np.all(uis[:, jj[0], jj[1]] == -1)
        ois = pack(data3d, fmask2d)
        assert ois.dtype == data3d.dtype
        assert np.all(ois == osoll)
        uis = unpack(ois, fmask2d)
        assert uis.dtype == data3d.dtype
        assert np.all(uis[:, ii[0], ii[1]] == data3d[:, ii[0], ii[1]])
        uis = unpack(ois, fmask2d, -1)
        jj = np.where(~mask2d)
        assert uis.dtype == data3d.dtype
        assert np.all(uis[:, ii[0], ii[1]] == data3d[:, ii[0], ii[1]])
        assert np.all(uis[:, jj[0], jj[1]] == -1)

        # data4d, mask2d
        ois = pack(data4d, mask2d)
        ii = np.where(mask2d)
        osoll = data4d[..., ii[0], ii[1]]
        assert ois.dtype == data4d.dtype
        assert np.all(ois == osoll)
        uis = unpack(ois, mask2d)
        assert uis.dtype == data4d.dtype
        assert np.all(uis[..., ii[0], ii[1]] == data4d[..., ii[0], ii[1]])
        uis = unpack(ois, mask2d, -1)
        jj = np.where(~mask2d)
        assert uis.dtype == data4d.dtype
        assert np.all(uis[..., ii[0], ii[1]] == data4d[..., ii[0], ii[1]])
        assert np.all(uis[..., jj[0], jj[1]] == -1)

        # data4d, mask3d
        ois = pack(data4d, mask3d)
        ii = np.where(mask3d)
        osoll = data4d[:, ii[0], ii[1], ii[2]]
        assert ois.dtype == data4d.dtype
        assert np.all(ois == osoll)
        uis = unpack(ois, mask3d)
        assert uis.dtype == data4d.dtype
        assert np.all(uis[:, ii[0], ii[1], ii[2]] == (
            data4d[:, ii[0], ii[1], ii[2]] ))
        uis = unpack(ois, mask3d, -1)
        jj = np.where(~mask3d)
        assert uis.dtype == data4d.dtype
        assert np.all(uis[:, ii[0], ii[1], ii[2]] == (
            data4d[:, ii[0], ii[1], ii[2]] ))
        assert np.all(uis[:, jj[0], jj[1], jj[2]] == -1)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
'''
    JAMS Eddy Covariance utilities

    Get help on each function by typing
    >>> help()
    help> eddybox.function
    Or
    >>> from jams import eddybox
    >>> help(eddybox.function)


    Provided functions (alphabetic w/o obsolete)
    ------------------
    eddycorr               Calculate time lags between wind and concentrations for EddyFlux.
    eddyspec               Performs spectrum analysis with EddySpec and SpecMean and determines inductances.
    energyclosure          Computes energy closure and correction for Eddy covaraince data
    fluxfill               Wrapper function for gapfill with file management and plotting.
    fluxflag               Quality flag calculation for Eddy Covariance data
    fluxpart               Wrapper function for nee2gpp including file management and plotting
    fluxplot               Plotting routine for Eddy Covariance or other ascii data file
    gapfill                Gapfill Eddy flux data.
    itc                    Calculation of integral turbulence characteristics after Thomas & Foken (2002)
    meteo4slt              EddyFlux supply with meteorological data.
    nee2gpp                Photosynthesis and ecosystem respiration from NEE Eddy flux data.
    nee2gpp_falge          nee2gpp using one fit for whole time period
    nee2gpp_lasslop        nee2gpp using the daytime method of Lasslop et al. (2010)
    nee2gpp_reichstein     nee2gpp using several fits as in Reichstein et al. (2005)
    planarfit              Planar fit of Eddy Covariance wind components
    profile2storage        Calculate storage fluxes from profile data to correct eddy data
    sltclean               Moves *.slt files in a deleted folder to exclude from processing (EddySoft files).
    spikeflag              Spike detection for Eddy Covariance data (and basically all other data)
    ustarflag              Friction velocity flagging for Eddy Covariance data


    Example
    -------
    see eddysuite.py


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python library,
    Department of Computational Hydrosystems, Helmholtz Centre for Environmental
    Research - UFZ, Leipzig, Germany.

    Copyright (c) 2014 Arndt Piayda, Matthias Cuntz - mc (at) macu (dot) de

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.


    History
    -------
    Written  AP, Sep 2014
'''
from .eddycorr          import eddycorr
from .eddyspec          import eddyspec
from .energyclosure     import energyclosure
from .fluxfill          import fluxfill
from .fluxflag          import fluxflag
from .fluxpart          import fluxpart
from .fluxplot          import fluxplot
from .gapfill           import gapfill
from .itc               import itc
from .meteo4slt         import meteo4slt
from .nee2gpp           import nee2gpp, nee2gpp_falge, nee2gpp_lasslop, nee2gpp_reichstein
from .planarfit         import planarfit
from .profile2storage   import profile2storage
from .sltclean          import sltclean
from .spikeflag         import spikeflag
from .ustarflag         import ustarflag

# Information
__author__   = "Arndt Piayda, Matthias Cuntz"
__version__  = '1.0'
__revision__ = "$Revision: 1796$"
__date__     = '$Date: 30.09.2014$'



#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
import numpy as np

def pet_oudin(temperature, lat, doy):
    """
        Calculates the daily potential evapotranspiration in [mm/d] following the Oudin formula.

        For the mathematical details of the PE formulation, see:
        Oudin, L., Hervieu, F., Michel, C., Perrin, C., Andreassian, V., Anctil, F. and Loumagne, C., 2005. 
        Which potential evapotranspiration input for a rainfall-runoff model? 
        Part 2 - Towards a simple and efficient PE model for rainfall-runoff modelling. 
        Journal of Hydrology 303(1-4), 290-306.

        For the calculation of extra-atmospheric global radiation, see
        Appendix C of the article by Morton, F.I., 1983. 
        Operational estimates of areal evapotranspiration and their significance to the science and 
        practice of hydrology. 
        Journal of Hydrology 66 (1/4), 1-76.


        Definition
        ----------
        def pet_oudin(temperature, lat, doy):


        Input
        -----
        temperature  (array of) temperature in degrees Celsius
        lat          (array of) latitudes in degrees N
        doy          (array of) day of the year (1 ... 366)


        Options
        -------
        none

        Output
        ------
        array with potential evapotranspiration (PET) based on Oudin formula in [mm/d]


        Restrictions
        ------------
        Shape of input arrays must be same


        Examples
        --------
        # PET based on Oudin
        >>> import numpy as np
        >>> temp = np.array([ 32.0,  32.0,  32.0,  32.0,       32.0,       32.0])
        >>> lat  = np.array([ 48.73, 48.73, 48.73, 9.27785325, 5.56671363, 5.56671363])
        >>> doy  = np.array([ 1,     2,     3,     364,        365,        366])
        >>> from autostring import astr
        >>> print(astr(pet_oudin(temp, lat, doy),3,pp=True))
        ['1.301' '1.309' '1.317' '4.844' '5.123' '5.127']


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2020 Juliane Mai - juliane.mai@uwaterloo.ca 

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  JM, Feb 2020
    """

    # check that inputs have all same shape
    ishape = np.shape(temperature)
    if not(np.shape(lat) == ishape):
        raise ValueError("pet_oudin: Array of latitudes must have same shape as temperature array")
    if not(np.shape(doy) == ishape):
        raise ValueError("pet_oudin: Array of day-of-year must have same shape as temperature array")

    # initialize
    pet_oudin = np.ones(ishape) * -9999.0

    teta  = 0.4093 * np.sin(doy/58.1 - 1.405)
    cosGz = np.maximum(0.001, np.cos(lat/57.3-teta))
    Gz    = np.arccos(cosGz)
    cosOM = np.maximum(-1.0, np.minimum(1.0, 1.0-cosGz/np.cos(lat/57.3)/np.cos(teta)))
    OM    = np.arccos(cosOM)
    Eta   = 1.0 + np.cos(doy/58.1)/30.0
    cosPz = cosGz + np.cos(lat/57.3)*np.cos(teta)*(np.sin(OM)/OM-1.0)

    # Global Radiation [W/m2]
    globrad = 446.0 * OM * Eta * cosPz

    # Potential Evapotranspiration [mm/d]
    pet_oudin = np.maximum(0.0, globrad/28.5 * (temperature+5.0)/100.)
    
    return pet_oudin

if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
import numpy as np


def tcherkez(Rstar, Phi=0.3, T=0.056,
             a2=1.0012, a3=1.0058, a4=1.0161,
             t1=0.9924, t2=1.0008, g=20e-3,
             RG=False, Rchl=False, Rcyt=False, fullmodel=True):
    """
        Calculates the Tcherkez model of 13C-discrimiantion in the Calvin cycle.


        Definition
        ----------
        def tcherkez(Rstar, Phi=0.3, T=0.056, a2=1.0012, a3=1.0058, a4=1.0161,
                     t1=0.9924, t2=1.0008, g=20e-3
                     RG=False, Rchl=False, Rcyt=False, fullmodel=True):


        Input
        -----
        Rstar      Isotope ratio of assimilated carbon, e.g. of Farquhar et al. (1982) model


        Optional Input
        --------------
        Phi       Vo/Vc: ratio of carboxylateion to oxygentation of Rubisco (default: 0.3)
        T         Relative flux of starch synthesis [mol(C6 of starch)/mol(CO2 assimilated)] (default: 0.056)
        a2        Inverse fractionation associated with aldolase
                  for the C-2 position of FBP (Fructose-1,6-bisphosphate) (default: 1.0012)
        a3        Same for C-3 of FBP (default: 1.0058)
        a4        Same for C-4 of FBP (default: 1.0161)
        t1        Inverse fractionation associated with trankelotase
                  for C-1 in E4P (erythrose-4-phosphate) and R5P (ribose-5-phosphate) (default: 0.9924)
        t2        Same for C-2 of X5P (xylulose-5-phosphate) (default: 1.0008)
        g         Isotope discrimination of photorespiratory decarboxylation of Gly (Glycine) (default: 20e-3)
        RG        If True, output isotope ratio of G3P (3-phosphoglyceraldehyde
                  or glyceraldehyde-3-phosphate) (default: False)
        Rchl      If True, output isotope ratio of chloroplastic hexoses and transitory starch (default: False)
        Rcyt      If True, output isotope ratio of cytoplasmic hexoses (default: False)
        fullmodel If True, output RG, Rchl and Rcyt (default: True)


        Output
        ------
        RG, Rchl, Rcyt    if fullmodel=True


        Restrictions
        ------------
        If at least one of RG, Rchl or Rcyt is given then fullmode=False.


        References
        ----------
        Tcherkez G, Farquhar GD, Badeck F & Ghashghaie J, Theoretical considerations about carbon isotope
            distribution in glucose of C3 plants, Functional Plant Biology 31, 857-877, 2004
        Gessler A, Tcherkez G, Peuke AD, Ghashghaie J & Farquhar GD, Experimental evidence for diel variations
            of the carbon isotope composition in leaf, stem and phloem sap organic matter in Ricinus communis,
            Plant, Cell and Environment 31, 941-953, 2008

        Examples
        --------
        >>> a       = -4.4e-3
        >>> b       = -27e-3
        >>> ca    = 353e-6
        >>> ci    = 0.7*ca
        >>> Delta = a+(b-a)*ci/ca
        >>> delta_a1 = -8e-3
        >>> Ra1      = (delta_a1+1.)*const.R13VPDB
        >>> Rstar1   = (1.-Delta)*Ra1
        >>> from autostring import astr
        >>> print(astr((np.array(tcherkez(Rstar1, Phi=0.3, T=0.056))/const.R13VPDB-1.)*1000.,3,pp=True))
        ['12.764' '17.125' '12.978']

        >>> delta_a2 = -7.8e-3
        >>> Ra2      = (delta_a2+1.)*const.R13VPDB
        >>> Rstar2   = (1.-Delta)*Ra2
        >>> R1 = (np.array(tcherkez([Rstar1, Rstar2], Rcyt=True))/const.R13VPDB-1.)*1000.
        >>> print(astr(R1,3,pp=True))
        [['12.978' '13.182']]

        >>> R1, R2 = tcherkez([Rstar1, Rstar2], Rchl=True, Rcyt=True)
        >>> print(astr((R1/const.R13VPDB-1.)*1000.,3,pp=True))
        ['17.125' '17.330']
        >>> print(astr((R2/const.R13VPDB-1.)*1000.,3,pp=True))
        ['12.978' '13.182']


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2012-2013 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Jan 2012
        Modified, MC, Feb 2013 - ported to Python 3
    """
    #
    if (RG | Rchl | Rcyt):
       fullmodel = False
    if fullmodel:
       RG   = True
       Rchl = True
       Rcyt = True
    #
    a2tilde = (1.+0.5*Phi-T) / ((2.*a2+1.)/3.+Phi*(2.*a2-0.5)/3.+T*(a2-2.))
    a3tilde = (1.+0.5*Phi-T) / ((2.*a3+1.)/3.+Phi*(2.*a3-0.5)/3.+T*(a3-2.))

    t1tilde = (1.+3.*T)/(t1+3.*T)*t1
    t2tilde = (1.+3.*T)/(t2+3.*T)*t2

    eps     = a3*a3tilde
    epsdash = (t1tilde+1.5*Phi)*a3*a3tilde/(3.*(1.+0.5*Phi-(1.+t2tilde)*a2*a2tilde/3.))

    iRG = np.array(Rstar) / (1.+Phi*(0.5-(1.+g)/(2.+g)*(eps+2.*a2*a2tilde*epsdash)/3.)+T*(a4-1.))

    iRchl = 1./6.*(epsdash*(1.+(a2*a2tilde*t2tilde)/t2)+eps*(2.+t1tilde/t1)+a4) * iRG
    iRcyt = 1./6.*(2.*eps+3.*(a2+1.)/(a2+2.)*epsdash*a2tilde+3.*a3tilde/(2.+a3)*(a3+2.*a4/(1.+a4))) * iRG

    out  = []
    if RG:
       out += [iRG]
    if Rchl:
       out += [iRchl]
    if Rcyt:
       out += [iRcyt]

    return out


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)




#!/usr/bin/env python
import numpy as np
from pyjams.jams.date2dec import date2dec


def fluxplot(infile, outfile, numheader=1, lineofvar=1, novalue=-9999,
             delimiter=',', format='ascii', format_str='%d-%m-%Y %H:%M',
             units=False, plot=False):
    '''
    Plotting routine for Eddy Covariance data generated by EddyFlux (and
    basically any other ascii data file) with first column containing time steps
    in either 'asci' or 'eng' time format. Loads the file and plots each column
    in a separate page of a pdf file. Missing values are plotted in red.


    Definition
    ----------
    fluxplot(infile, outfile, numheader=1, lineofvar=1, novalue=-9999,
             format='ascii', format_str_x='%d-%m-%Y %H:%M', units=False,
             plot=False):


    Input
    -----
    infile      str, path and name of the input file
    outfile     str, path and name of the output pdf


    Optional Input
    --------------
    numhead     int, number of header lines in the input files (default: 1)
    lineofvar   int, line in input file containing variable names (default: 1)
    novalue     int/float, missing value in input file (default: -9999)
    delimiter   str, delimiter of the input file
    format      str, format of time stamps in input files. 'ascii' or 'eng' is
                possible (default: 'ascii')
    format_str  str, format string for plot date formatting
                (default: '%d-%m-%Y %H:%M')
    units       bool, if True, all units of a standard EddyFlux output file
                are added to each plot. Needs to be set False if other files
                are plotted (default: False)
    plot        bool, only if True, the plot is performed (default: False)


    Output
    ------
    outfile     pdf with each page containing a plot of one column of the input
                file


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python library,
    Department of Computational Hydrosystems, Helmholtz Centre for Environmental
    Research - UFZ, Leipzig, Germany.

    Copyright (c) 2014 Arndt Piayda

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.


    History
    -------
    Written,  AP, Aug 2014
    '''

    if plot:
        import matplotlib
        matplotlib.use('PDF')
        import matplotlib.pyplot as plt
        import matplotlib.ticker as ticker
        from matplotlib.backends.backend_pdf import PdfPages
        from matplotlib import rcParams
    else:
        raise ValueError('fluxplot: plot must be True')

    ############################################################################
    # set parameters for matplotlib
    rcParams['ps.papersize']        = 'a4'
    rcParams['savefig.orientation'] = 'landscape'
    rcParams['lines.linewidth']     = 0.75
    rcParams['axes.linewidth']      = 1
    rcParams['axes.titlesize']      = 12
    rcParams['axes.labelsize']      = 8
    rcParams['xtick.labelsize']     = 8
    rcParams['ytick.labelsize']     = 8
    rcParams['legend.fontsize']     = 8
    rcParams['path.simplify']       = False

    ############################################################################
    # reading input file
    d       = np.loadtxt(infile, dtype='|S100', delimiter=delimiter)
    header  = d[lineofvar-1,1:]
    date    = d[numheader:,0]
    value   = d[numheader:,1:]
    value[value==''] = str(novalue)
    value   = value.astype(np.float)
    missing = np.where(value == novalue, 0, np.NaN)
    value   = np.where(value == novalue, np.NaN, value)
    length  = len(date)
    if units:
        units = ['W m-2', 'W m-2', 'mmol m-2 s-1', 'mumol m-2 s-1', 'm s-1',
                 'kg m-1 s-2', 'm2 s-2', 'm2 s-2', 'm2 s-2', 'K2', 'mmol2 m-6',
                 'mmol2 m-6', 'm s-1', 'm s-1', 'm s-1', ' deg C', 'mmol m-3',
                 'mmol m-3', 'm s-1', 'deg', 'deg', 'deg', 'deg', 'samples',
                 'samples', 'm', '-', 'deg C', 'W m-2', 'W m-2', 'mmol m-2 s-1',
                 'mumol m-2 s-1', 'm', 'm', 'm', '-', '-', '-', '-', '-', '-',
                 'm2 s-2', 'm2 s-2', 'm2 s-2', '-', 'samples', 'samples',
                 'samples', 'samples', 'samples', 'samples', 'samples',
                 'samples', 'samples', 'samples', 'samples', 'samples',
                 'samples', 'hPa', 'deg C', '%', 'hPa', 'kg m-3', 'samples',
                 'samples', 'm', 'm']
    else:
        units = ['']*length
    ############################################################################
    # convert input date in JD and array format
    date01 = date2dec(yr=1, mo=1, dy=2, hr=0, mi=0, sc=0)
    if format=='ascii':
        jdarr = date2dec(ascii=date) - date01
    elif format=='eng':
        jdarr = date2dec(eng=date) - date01
    else:
        raise ValueError('fluxplot: format unknown')

    ############################################################################
    # plot
    majticks = matplotlib.dates.MonthLocator(bymonthday=1)
    pdf_pages = PdfPages(outfile)

    ############################################################################
    for i in xrange(len(header)):#len(header)

        fig = plt.figure(i)
        ax = fig.add_subplot(111)
        ax.plot(jdarr, value[:,i], 'b-')
        ylim = ax.get_ylim()
        if (ylim[0] < 0) and (ylim[1] > 0):
            ax.axhline(y=0, linewidth=0.75, color='k')
            ax.plot(jdarr, missing[:,i], 'r-', lw=2)
        else:
            ax.plot(jdarr, np.where(missing[:,i] == 0, ylim[0], np.NaN), 'r-', lw=2)

        ax.set_xlabel('Date [DD-MM-YYYY]')
        ylabel = '%s \n %s' %(header[i], units[i])
        ax.set_ylabel(ylabel)

        ax.set_xlim(jdarr[0],jdarr[-1])
        ax.xaxis.set_major_locator(majticks)
        ax.xaxis.set_major_formatter(matplotlib.dates.DateFormatter(format_str))
        fig.autofmt_xdate()

        pdf_pages.savefig(fig)
        plt.close()
    pdf_pages.close()

if __name__ == '__main__':
    import doctest
    doctest.testmod()



#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
import numpy as np
from pyjams.jams.around import around
from pyjams.const import eps


__all__ = ['yrange']


def yrange(*args, **kwargs):
    """
    Calculates plot range from input array


    Definition
    ----------
    def yrange(arr, symmetric=False):


    Input
    -----
    arr        number array


    Optional Input
    --------------
    symmetric  if True, range will be symmetric around 0.
               if min(arr)<0 and max(arr)>0.


    Output
    ------
    Range to be used as [xyz]range


    Restrictions
    ------------
    Uses around.
    Does not work well for 0<range<1. Use yrange(arr*10.)/10.


    Examples
    --------
    >>> import numpy as np
    >>> print(yrange(range(102)))
    [0.0, 101.0]

    >>> print(yrange(np.arange(102)-10.))
    [-10.0, 91.0]

    >>> print(yrange(np.arange(102)-10., symmetric=True))
    [-91.0, 91.0]

    >>> print(yrange(range(102), range(1002), np.arange(10002)))
    [0.0, 10001.0]

    >>> print(yrange(-np.arange(102), np.arange(1002), np.arange(10002)))
    [-101.0, 10001.0]

    >>> print(yrange(-np.arange(102), np.arange(1002), np.arange(10002), symmetric=True))
    [-10001.0, 10001.0]

    >>> print(yrange(-np.arange(102), np.arange(1002), np.arange(10002), symmetric=False))
    [-101.0, 10001.0]

    >>> a = np.ma.arange(102)
    >>> a[-1] = np.ma.masked
    >>> print(yrange(a))
    [0.0, 100.0]

    >>> a = np.ma.arange(102, dtype=float)
    >>> a[-1] = np.ma.masked
    >>> a[-2] = np.nan
    >>> print(yrange(a))
    [0.0, 99.0]


    License
    -------
    This file is part of the JAMS Python package, distributed
    under the MIT License. The JAMS Python package originates from the former
    UFZ Python library, Department of Computational Hydrosystems, Helmholtz
    Centre for Environmental Research - UFZ, Leipzig, Germany.

    Copyright (c) 2012-2016 Matthias Cuntz - mc (at) macu (dot) de

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.


    History
    -------
    Written,  Matthias Cuntz, Jan 2012
    Modified, Matthias Cuntz, Feb 2013 - ported to Python 3
              Matthias Cuntz, Nov 2013 - accept more than 1 array as input
              Matthias Cuntz, Nov 2013 - masked arrays
              Matthias Cuntz, Apr 2014 - assert
              Matthias Cuntz, Nov 2016 - const.tiny -> const.eps
              Matthias Cuntz, Nov 2019 - mask NaN values, e.g. from Pandas
    """
    # Check input
    assert len(args) > 0, 'no input argument given.'
    counter = 0
    for i in args:
        if np.any(np.isnan(i)):
            arr = np.ma.array(i, mask=np.isnan(i))
        else:
            arr    = np.ma.array(i)
        minarr = np.ma.amin(arr)
        maxarr = np.ma.amax(arr)
        # Round to max difference between adjacent values
        if arr.size == 1:
            expom = 0
        else:
            sarr    = np.ma.sort(arr)
            maxdiff = np.ma.amax(np.diff(sarr))
            expom   = np.ma.log10(maxdiff)
            if expom > 0:
                expom = int(np.floor(expom + 10. * eps * 10.))
            else:
                expom = int(np.floor(expom - 10. * eps))
        if counter == 0:
            minall   = minarr
            maxall   = maxarr
            expomall = expom
        else:
            minall   = np.minimum(minall, minarr)
            maxall   = np.maximum(maxall, maxarr)
            if expomall > 0:
                expomall = np.maximum(expomall, expom)
            else:
                expomall = np.minimum(expomall, expom)
        counter += 1

    # Round range
    mini = around(minall, expomall, floor=True)
    maxi = around(maxall, expomall, ceil=True)

    if 'symmetric' in kwargs:
        if kwargs['symmetric']:
            if (mini * maxi < 0.):
                maxmax = np.maximum(np.abs(mini), np.abs(maxi))
                maxi   = maxmax
                mini   = -maxmax

    # Return range
    return [mini, maxi]


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # import numpy as np

    # print(yrange(np.arange(102)))
    # #[0.0, 101.0]

    # print(yrange(np.arange(102)-10.))
    # #[-10.0, 91.0]

    # print(yrange(np.arange(102)-10., symmetric=True))
    # #[-91.0, 91.0]

    # print(yrange(np.arange(102), np.arange(1002), np.arange(10002)))
    # #[0.0, 10001.0]

    # print(yrange(-np.arange(102), np.arange(1002), np.arange(10002)))
    # #[-101.0, 10001.0]

    # print(yrange(-np.arange(102), np.arange(1002), np.arange(10002), symmetric=True))
    # #[-10001.0, 10001.0]

    # print(yrange(-np.arange(102), np.arange(1002), np.arange(10002), symmetric=False))
    # #[-101.0, 10001.0]

    # a = np.ma.arange(102)
    # a[-1] = np.ma.masked
    # print(yrange(a))
    # #[0.0, 100.0]



#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
'''
Get IsoGSM2 output for one or several lat,lon.

--------------------------------------------------------
usage: get_isogsm2.py [-h] [-l lat,lon] [-o] [-p path] [-u url] [latlon_file]

Get IsoGSM2 output for one or several lat,lon.

positional arguments:
  latlon_file           File with lat lon info (only these two header columns
                        must begin with lat and lon (case-insensitive). File
                        only used if no -l lat,lon given.

optional arguments:
  -h, --help            show this help message and exit
  -l lat,lon, --latlon lat,lon
                        latitude,longitude to extract from IsoGSM2 (default:
                        read from input file). If latitude or longitude is
                        negative, use the notation -l=lat,lon instead of -l
                        lat,lon.
  -o, --override        Do not check that output file already exists. Override
                        existing output file (default: False).
  -p path, --path path  Output directory (default: current directory ".").
  -u url, --url url     URL of IsoGSM2 base directory (default:
                        http://isotope.iis.u-tokyo.ac.jp/~kei/tmp/isogsm2).


Example
-------
# Hesse
python get_isogsm2.py -l 48.6742166667,7.06461666667

# Lucas site collection
python get_isogsm2.py Leaf_water_data_combined_v4.csv


History
-------
Written,  Matthias Cuntz, Oct 2018
Modified, Matthias Cuntz, Nov 2018
              - make it callable function as well as script.
          Matthias Cuntz, Jan 2021
              - (almost) flake8 compatible
'''
import os
import glob
import fileinput
import numpy as np
# Reading from HTTP
try:
    isrequests = True
    import requests_not_used
except ImportError:  # no requests module
    isrequests = False
    try:                 # Python 3
        import urllib.request as liburl
    except ImportError:  # Python 2
        import urllib as liburl
# Parsing HTML pages
try:
    from BeautifulSoup import BeautifulSoup
except ImportError:
    from bs4 import BeautifulSoup
try:
    bparser = 'lxml'
    import lxml
except:
    bparser = None
# Reading user input
try:               # Python 2
    input = raw_input
except NameError:  # Python 3
    pass


__all__ = ['get_isogsm2']


isogsm2base = 'http://isotope.iis.u-tokyo.ac.jp/~kei/tmp/isogsm2'
# IsoGSM2 T62 grid
isogsm2lons = np.arange(192) * 1.875
isogsm2lats = np.array(
    [-88.542, -86.653, -84.753, -82.851, -80.947, -79.043, -77.139, -75.235,
     -73.331, -71.426, -69.522, -67.617, -65.713, -63.808, -61.903, -59.999,
     -58.094, -56.189, -54.285, -52.380, -50.475, -48.571, -46.666, -44.761,
     -42.856, -40.952, -39.047, -37.142, -35.238, -33.333, -31.428, -29.523,
     -27.619, -25.714, -23.809, -21.904, -20.000, -18.095, -16.190, -14.286,
     -12.381, -10.476, -8.571, -6.667, -4.762, -2.857, -0.952,
     0.952, 2.857, 4.762, 6.667, 8.571, 10.476, 12.381, 14.286, 16.190, 18.095,
     20.000, 21.904, 23.809, 25.714, 27.619, 29.523, 31.428, 33.333, 35.238,
     37.142, 39.047, 40.952, 42.856, 44.761, 46.666, 48.571, 50.475, 52.380,
     54.285, 56.189, 58.094, 59.999, 61.903, 63.808, 65.713, 67.617, 69.522,
     71.426, 73.331, 75.235, 77.139, 79.043, 80.947, 82.851, 84.753, 86.653,
     88.542])


# -------------------------------------------------------------------------
# Functions
#

# Read URL
def get_url(url, **kwargs):
    ''' Returns the content of website or data file as string '''
    if isrequests:
        rr = requests.get(url, **kwargs)
        return rr.text
    else:
        if 'proxies' in kwargs:
            if kwargs['proxies'] is not None:
                proxy  = liburl.ProxyHandler(proxies)
                auth   = liburl.HTTPBasicAuthHandler()
                opener = liburl.build_opener(proxy, auth, liburl.HTTPHandler)
                liburl.install_opener(opener)
        rr = liburl.urlopen(url)
        return rr.read().decode('utf-8')


# Read URL, write to file
def url2filename(url, ifile, **kwargs):
    ''' Returns file with the local name "ifile" from an
        url such as http://website/file.dat '''
    rr = get_url(url, **kwargs)

    ff = open(ifile, 'w')
    print(rr, file=ff)
    ff.close()

    return


# Read URL, write to file with filename from URL
def url2file(url, **kwargs):
    ''' Returns file from an url such as http://website/file.dat
        having the same local name as the remote name '''
    if url[-1] == '/':
        raise ValueError('url does not end with valid filename.')

    ifile = url[url.rfind('/') + 1:]
    url2filename(url, ifile, **kwargs)

    return


# Get IsoGSM2 available years
def get_years(url, **kwargs):
    ''' Get all the years from IsoGSM2 base website '''
    page = get_url(url, **kwargs)

    bpage = BeautifulSoup(page, features=bparser)

    # '/~kei/tmp/', '1979/', ...
    href = [ i.get('href') for i in bpage.findAll('a') ]
    # '/~kei/tmp', '1979', ...
    hdir = [ i[:-1] for i in href if i.endswith('/') ]
    # 1979, 1980, ...
    yrs = []
    for hh in hdir:
        try:
            yrs.append(int(hh))
        except:
            pass
    return yrs


# Get IsoGSM2 files in year
def get_filenames(url, **kwargs):
    ''' Get all files in yearly directory of IsoGSM2 '''
    page = get_url(url, **kwargs)

    bpage = BeautifulSoup(page, features=bparser)

    # '/~kei/tmp/isogsm2/', 'x001y001_isogsm2_6hrly_2013.dat', ...
    href = [ i.get('href') for i in bpage.findAll('a') ]
    # 'x001y001_isogsm2_6hrly_2013.dat', ...
    # files = [ i for i in href if i.startswith('x') ]
    # 'x001y001_isogsm2_6hrly_2013.dat', ...
    files = [ i for i in href
              if (not i.startswith('?')) and (not i.endswith('/')) ]
    return files


# Get lat lon in IsoGSM2 grid
def closest(vec, num, value=False):
    ''' Index in array at which the entry is closest to a given number. '''
    out = np.argmin(np.abs(np.array(vec) - num))
    if value:
        return vec.flat[out]
    else:
        return out


def lonlat2xy(ilon, ilat):
    ''' Get x and y coordinates in IsoGSM2 grid of given lat and lon. '''
    if ilon < 0:
        ilon += 360.  # IsoGSM2 grid is 0-360.
    x = closest(isogsm2lons, ilon) + 1
    y = closest(isogsm2lats, ilat) + 1
    return (x, y)


# -------------------------------------------------------------------------
# Main routine
#

def get_isogsm2(latlon, baseurl=isogsm2base, path='.', override=False):
    '''Get IsoGSM2 output.

    Downloads all available years of IsoGSM2 output for a specific lat,lon,
    a list of lats and lons, or lat/lon read from columns of an input file.

    If override=False, checks if the data is already available in the
    download directory.
    It expects files with the same naming convention than its own, i.e.
        x, y = lonlat2xy(lon, lat)
        ofile = 'lat{:.3f}lon{:.3f}_isogsm2_6hrly_{:s}-{:s}.dat'.format(
                 isogsm2lats[y-1], isogsm2lons[x-1],
                 str(yrs[0]), str(yrs[-1]))
    It checks if the same lat/lon is present in the download directory by
    checking ONLY filenames.


    Definition
    ----------
    def get_isogsm2(latlon, baseurl=isogsm2base, path='.', override=False):


    Input
    -----
    latlon    string or list/tuple/ND-array
              latlon can be
                  string of one 'lat,lon', e.g. 48.5,7,
                  list, tuple or nd-array with iterables of lats and lons, e.g.
                      [[48.5,49],[7,15]], ([48.5,49],[7,15]),
                      (3,2)-array: np.array([[48.5,49,50],[7,15,-10]])
                  filename of file wherein two columns in the first header line
                  start with strings 'lat' and 'lon' (case-insensitive).
                  Only these two header columns must begin with lat and lon.


    Parameters
    ----------
    baseurl     string
                URL of IsoGSM2 base directory
                default: http://isotope.iis.u-tokyo.ac.jp/~kei/tmp/isogsm2
    path        string (default: '.')
                Output path
    override    boolean (default: False)
                If True, check if data is already available in path.
                Expect files with the naming convention
                x, y = lonlat2xy(lon, lat)
                ofile = 'lat{:.3f}lon{:.3f}_isogsm2_6hrly_{:s}-{:s}.dat'.format(
                    isogsm2lats[y-1], isogsm2lons[x-1],
                    str(yrs[0]), str(yrs[-1]))
                Check if lat/lon was already downloaded in path by checking
                ONLY filenames.


    Ouput
    -----
    Returns filenames of the output files that were newly written or already
    present in path:
       'lat{:.3f}lon{:.3f}_isogsm2_6hrly_{:s}-{:s}.dat'.format(
           isogsm2lats[y-1], isogsm2lons[x-1], str(yrs[0]), str(yrs[-1]))


    Examples
    --------
    >>> latlon = '48.6742166667,7.06461666667'
    >>> ff = get_isogsm2(latlon)
    >>> import os
    >>> file1 = 'lat48.571lon7.500_isogsm2_6hrly_1979-2017.dat'
    >>> if not os.path.exists(file1): print('No file: ', file1)


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License.

    Copyright (c) 2018-2021 Matthias Cuntz - mc (at) macu (dot) de

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.


    History
    -------
    Written,  Matthias Cuntz, Oct 2018
    Modified, Matthias Cuntz, Nov 2018
                  - make it callable function as well as script.
              Matthias Cuntz, Jan 2021
                  - (almost) flake8 compatible
    '''
    # get lats and lons
    if isinstance(latlon, str):
        if ',' in latlon:
            ilat, ilon = [ float(i) for i in latlon.split(',') ]
            lats = [ilat]
            lons = [ilon]
        else:
            if not os.path.exists(latlon):
                raise IOError(f'Did not find latlon: {latlon}. latlon must'
                              f' be either a file name or a string "lat,lon".')
            lats = []
            lons = []
            ff = open(latlon, 'r')
            lhead = ff.readline().rstrip()
            for delim in [',', ';', None]:
                head  = lhead.split(delim)
                nhead = len(head)
                if nhead > 1:
                    break
            if nhead == 1:
                ff.close()
                raise IOError(f'Cannot detect delimter in latlon file:'
                              f' {latlon}')
            ii = [ i for i in range(nhead)
                   if head[i].lower().startswith('lat') ]
            if len(ii) != 1:
                ff.close()
                raise IOError(f'Found no or more than one column starting'
                              f' with "lat" in latlon file: {latlon}')
            iilat = ii[0]
            ii = [ i for i in range(nhead)
                   if head[i].lower().startswith('lon') ]
            if len(ii) != 1:
                ff.close()
                raise IOError(f'Found no or more than one column starting'
                              f' with "lon" in latlon file: {latlon}')
            iilon = ii[0]
            for line in ff:
                s  = line.rstrip()
                ss = s.split(delim)
                lats.append(ss[iilat])
                lons.append(ss[iilon])
            ff.close()
            if (len(lats) == 0) or (len(lons) == 0):
                raise IOError(
                    'Could not get lat,lon list from latlon file: ' +
                    latlon)
    elif isinstance(latlon, (tuple, list, np.ndarray)):
        if isinstance(latlon, (tuple, list)):
            lats = latlon[0]
            lons = latlon[1]
        else:
            lats = latlon[0, :]
            lons = latlon[1, :]
    else:
        raise IOError('latlon must be either a string "lat,lon", a filename,'
                      ' or an iterable with lats and lons.')
    nlat = len(lats)

    # make output directory
    if not os.path.exists(path):
        os.makedirs(path)

    # Check existing files
    hasfile  = [False] * nlat
    isofiles = [''] * nlat
    files = glob.glob(path + '/lat*.dat')
    if (len(files) > 0) and (not override):
        for ll in range(nlat):
            x, y = lonlat2xy(float(lons[ll]), float(lats[ll]))
            ilat = '{:.3f}'.format(isogsm2lats[y - 1])
            ilon = '{:.3f}'.format(isogsm2lons[x - 1])
            for ff in files:
                fs = os.path.basename(ff).split('_')[0][3:]
                lat, lon = fs.split('lon')
                if ((lat == ilat) and (lon == ilon)):
                    hasfile[ll]  = True
                    isofiles[ll] = ff
    # Return if all exist
    if np.all(hasfile):
        return isofiles

    # Detect IsoGSM2 years
    try:
        yrs = get_years(baseurl)
    except Exception as e:
        if str(e) == 'HTTP Error 407: Proxy Authentication Required':
            print('Enter your http proxy details:'
                  ' http://user:password@proxy_server:port')
            print('For example: http://mcuntz:password@revelec.inra.fr:3128')
            user = input('User (mcuntz):')
            if user == '':
                user = 'mcuntz'
            print('Password: ', end='', flush=True)
            import getpass
            password = getpass.getpass()
            proxy_server = input('Proxy server (revelec.inra.fr):')
            if proxy_server == '':
                proxy_server = 'revelec.inra.fr'
            port = input('Port (3128):')
            if port == '':
                port = '3128'
            proxy_string  = 'http://' + str(user) + ':' + str(password)
            proxy_string += '@' + str(proxy_server) + ':' + str(int(port))
            proxies = {"http": proxy_string}
            # needs proxies only once during session
            yrs = get_years(baseurl, proxies=proxies)
        else:
            raise IOError(str(e))

    for ll in range(nlat):
        if not hasfile[ll]:
            print('  Get lat lon: ', lats[ll], lons[ll])
            x, y = lonlat2xy(float(lons[ll]), float(lats[ll]))
            llfiles = []
            for yy in yrs:
                iifile = 'x{:03d}y{:03d}_isogsm2_6hrly_{:s}.dat'.format(
                    x, y, str(yy))
                url = baseurl + '/' + str(yy) + '/' + iifile
                print('    ', url)
                try:
                    url2file(url)
                    llfiles.append(path + '/' + iifile)
                except:
                    print('Could not find URL: ', url)

            # concatenate files
            isofiles[ll] = (
                path + '/' +
                'lat{:.3f}lon{:.3f}_isogsm2_6hrly_{:s}-{:s}.dat'.format(
                    isogsm2lats[y - 1], isogsm2lons[x - 1], str(yrs[0]),
                    str(yrs[-1])))
            print('  Write ', isofiles[ll])
            with open(isofiles[ll], 'w') as fout:
                fin = fileinput.input(llfiles)
                for line in fin:
                    # write header only of first file and
                    # strip emtpy lines at end of files
                    if ( ((fileinput.filelineno() > 2) or
                          (fileinput.lineno() < 3)) and
                         line.strip() != '' ):
                        fout.write(line)
                fin.close()

            # clean downloaded files
            for ll in llfiles:
                os.remove(ll)

    return isofiles


# -------------------------------------------------------------------------
# Command line arguments
#

def main(iargs):

    import argparse

    latlon   = None
    override = False
    path     = '.'
    baseurl  = isogsm2base
    parser  = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description='''Get IsoGSM2 output for one or several lat,lon.''')
    parser.add_argument(
        '-l', '--latlon', action='store',
        default=latlon, dest='latlon', metavar='lat,lon',
        help=('latitude,longitude to extract from IsoGSM2 (default: read'
              ' from input file). If latitude or longitude is negative,'
              ' use the notation -l=lat,lon instead of -l lat,lon.'))
    parser.add_argument(
        '-o', '--override', action='store_true', default=override,
        dest='override',
        help=('Do not check that output file already exists. '
              'Override existing output file (default: False).'))
    parser.add_argument(
        '-p', '--path', action='store', default=path, dest='path',
        metavar='path',
        help='Output directory (default: current directory ".").')
    parser.add_argument(
        '-u', '--url', action='store',
        default=baseurl, dest='baseurl', metavar='url',
        help='URL of IsoGSM2 base directory (default: ' + baseurl + ').')
    parser.add_argument(
        'ifile', nargs='?', default=None, metavar='latlon_file',
        help=('File with lat lon info (only these two header columns must'
              ' begin with lat and lon (case-insensitive). File only used'
              ' if no -l lat,lon given.'))

    args     = parser.parse_args(iargs)
    latlon   = args.latlon
    override = args.override
    path     = args.path
    baseurl  = args.baseurl
    ifile    = args.ifile

    del parser, args

    # Check input
    if (latlon is None) and (ifile is None):
        raise IOError('-l lat,lon or input file must be given.')

    if (latlon is not None):
        ilatlon = latlon
    else:
        ilatlon = ifile

    import time as ptime
    t1 = ptime.time()

    ofiles = get_isogsm2(ilatlon, baseurl=baseurl, path=path,
                         override=override)
    print('Files: ', ofiles)

    t2    = ptime.time()
    strin = ('[m]: {:.1f}'.format((t2 - t1) / 60.) if (t2 - t1) > 60. else
             '[s]: {:d}'.format(int(t2 - t1)))
    print('Time elapsed', strin)



#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
import numpy as np
"""
    Defines common error measures.


    Definition
    ----------
    def bias(obs,mod):      bias
                            mean(obs) - mean(mod)
    def mae(obs,mod):       mean absolute error
                            mean(abs(obs - mod))
    def mse(obs,mod):       mean square error
                            mean((obs - mod)**2)
    def rmse(obs,mod):      root mean square error
                            sqrt(mean((obs - mod)**2))
    def nse(obs,mod):       Nash-Sutcliffe efficiency
                            1 - sum((obs - mod)**2) / sum((obs - mean(obs))**2)
    def kge(obs,mod):       Kling-Gupta efficiency
                            1 - sqrt((1-r)**2 + (1-a)**2 + (1-b)**2),
                            where r is the Pearson correlation of obs and mod,
                                  a is mean(mod) / mean(obs), and
                                  b is std(mod) / std(obs)
    def pearson(obs,mod):   Pearson's correlation coefficient
                            mean((obs-mean(obs))/stddev(obs) * (mod-mean(mod))/stddev(mod))


    Input
    -----
    obs        ND-array
    mod        ND-array


    Output
    ------
    Measure calculated along the first axis.


    Restrictions
    ------------
    None


    Examples
    --------
    >>> # Create some data
    >>> obs = np.array([12.7867, 13.465, 14.1433, 15.3733, 16.6033])
    >>> mod = np.array([12.8087, 13.151, 14.3741, 16.2302, 17.9433])

    >>> # bias
    >>> print(np.round(bias(obs, mod),2))
    -0.43

    >>> print(np.round(mae(obs, mod),2))
    0.55

    >>> print(np.round(mse(obs, mod),2))
    0.54

    >>> print(np.round(rmse(obs, mod),2))
    0.73

    >>> print(np.round(nse(obs, mod),2))
    0.71

    >>> print(np.round(kge(obs, mod),2))
    0.58

    >>> print(np.round(pear2(obs, mod),2))
    0.99

    >>> print(np.round(confint(obs, p=0.95)))
    [ 13.  16.]


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python library,
    Department of Computational Hydrosystems, Helmholtz Centre for Environmental
    Research - UFZ, Leipzig, Germany.

    Copyright (c) 2014-2016 Arndt Piayda, Matthias Cuntz, Stephan Thober - mc (at) macu (dot) de

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.


    History
    -------
    Written, AP, Jul 2014
    Modified MC, Dec 2014 - use simple formulas that work with normal and masked arrays but do not deal with NaN
    Modified AP, Sep 2015 - add confidence interval
    Modified ST, Nov 2015 - added KGE
    Modified MC, May 2016 - calc along 1st axis so that measures work for ND-arrays
"""

all = ['bias', 'mae', 'mse', 'rmse', 'nse', 'kge', 'pearson']

def bias(obs, mod):
    """
    Bias = mean(obs) - mean(mod)
    """
    return obs.mean(axis=0) - mod.mean(axis=0)


def mae(obs, mod):
    """
    Mean absolute error = mean(abs(obs - mod))
    """
    assert cmp(obs.shape, mod.shape) == 0, 'Shapes of two input arrays do not match.'
    return (np.ma.abs(obs-mod)).mean(axis=0)


def mse(obs,mod):
    """
    Mean squared error = mean((obs - mod)**2)
    """
    assert cmp(obs.shape, mod.shape) == 0, 'Shapes of two input arrays do not match.'
    return ((obs-mod)**2).mean(axis=0)


def rmse(obs,mod):
    """
    Root mean squared error = sqrt(mean((obs - mod)**2))
    """
    assert cmp(obs.shape, mod.shape) == 0, 'Shapes of two input arrays do not match.'
    return ((obs-mod)**2).mean(axis=0)**0.5


def nse(obs,mod):
    """
    Nash-Sutcliffe-Efficiency = 1 - sum((obs - mod)**2) / sum((obs - mean(obs))**2)
    """
    return 1. - ((obs-mod)**2).sum(axis=0)/((obs-obs.mean(axis=0))**2).sum(axis=0)


def kge(obs,mod):
    """
    Kling-Gupta-Efficiency = 1 - sqrt((1-r)**2 + (1-a)**2 + (1-b)**2),
    where r is the Pearson correlation of obs and mod,
          a is mean(mod) / mean(obs), and
          b is std(mod) / std(obs)
    """
    return 1. - np.sqrt((1.-pearson(obs, mod))**2 + (1.-mod.mean(axis=0)/obs.mean(axis=0))**2 + (1.-mod.std(axis=0, ddof=1)/obs.std(axis=0, ddof=1))**2)


def pearson(obs,mod):
    """
    Pearson's correlation coefficient = mean((x-mean(x))/stddev(x) * (y-mean(y))/stddev(y))
    """
    return ((obs-obs.mean(axis=0))/obs.std(axis=0)*(mod-mod.mean(axis=0))/mod.std(axis=0)).mean(axis=0)


if __name__ == '__main__':
    import doctest
    doctest.testmod()



#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
import numpy as np

def in_poly(P, coord_x, coord_y):
    """
        Determines whether a 2D point falls within a polygon, on a vertex or
        edge, or is located outside of a polygon. The polygon can be convex or
        not.


        Definition
        ----------
        def in_poly(P, coord_x, coord_y):


        Input
        -----
        P         2D list or np.array,
                  x and y coordinates of the point in question in the form [x,y]
        coord_x   np.array, x coordinates of the polygon
        coord_y   np.array, y coordinates of the polygon


        Output
        ------
        integer, 1 = point inside polygon
                 0 = point on vertex/edge
                -1 = point outside polygon


        Restrictions
        ------------
        The method is only applicable for 2D polygons and points.


        References
        ----------
        This routine is re-coded from the JAMS Fortran package.
        Copyright: Juliane Mai, 2012.
        
        The original version of the source code (pnpoly) was implemented by
        W. Randolph Franklin. It had been, however, assigning insufficiently
        vertex/edge points.


        Example
        --------
        >>> coord_x = np.array([2.,2.,5.,7.,5.])
        >>> coord_y = np.array([1.,4.,6.,3.,1.])

        # point inside polygon
        >>> P = [4.,3.]
        >>> print(in_poly(P, coord_x, coord_y))
        1

        # point outside polygon
        >>> P = [8.,6.]
        >>> print(in_poly(P, coord_x, coord_y))
        -1

        # point on edge/vertex of polygon
        >>> P = [2.,2.]
        >>> print(in_poly(P, coord_x, coord_y))
        0


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2012-2014 Arndt Piayda, Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  AP, Nov 2012
        Modified, MC, Nov 2012 - documentation change, return 0 immediately
                  MC, Feb 2013 - ported to Python 3
                  MC, Oct 2013 - inpoly
                  MC, Apr 2014 - assert
    """

    # ironing :-)
    coord_x, coord_y = coord_x.flatten(), coord_y.flatten()

    # test input sizes
    assert np.size(coord_x) == np.size(coord_y), 'in_poly: coord_x and coord_y must have same size.'

    n = coord_x.size
    # result is outside as long as no other test works
    erg = -1

    # relative coordinates
    X = coord_x - P[0]
    Y = coord_y - P[1]

    # Edge test
    if np.any((X==0.) & (Y==0.)): return 0

    for i in range(n):
        # vertical Vertex test
        j = (i+1) % n
        if (coord_x[i] == coord_x[j]) and (coord_x[i] == P[0]):
            ly = (P[1]-coord_y[j]) / (coord_y[i]-coord_y[j])
            if (ly >= 0.) and (ly <= 1.): return 0

        # horizontal Vertex test
        if (coord_y[i] == coord_y[j]) and (coord_y[i] == P[1]):
            lx = (P[0]-coord_x[j]) / (coord_x[i]-coord_x[j])
            if (lx >= 0.) and (lx <= 1.): return 0

        # Inside test
        MX = X[i] >= 0.
        NX = X[j] >= 0.
        MY = Y[i] >= 0.
        NY = Y[j] >= 0.

        test1 = not((MY or NY) and (MX or NX)) or (MX and NX)
        test2 = not(MY and NY and (MX or NX) and not(MX and NX))

        if (not test1):
            if test2:
                tt = (Y[i]*X[j] - X[i]*Y[j]) / (X[j] - X[i])
                if tt == 0.:
                    return 0
                elif tt > 0.:
                    erg = -erg
            else:
                erg = -erg
    return erg


def inpoly(*args, **kwargs):
    """
        wrapper for in_poly
        def in_poly(P, coord_x, coord_y):


        Example
        --------
        >>> coord_x = np.array([2.,2.,5.,7.,5.])
        >>> coord_y = np.array([1.,4.,6.,3.,1.])

        # point inside polygon
        >>> P = [4.,3.]
        >>> print(inpoly(P, coord_x, coord_y))
        1

        # point outside polygon
        >>> P = [8.,6.]
        >>> print(inpoly(P, coord_x, coord_y))
        -1

        # point on edge/vertex of polygon
        >>> P = [2.,2.]
        >>> print(inpoly(P, coord_x, coord_y))
        0
    """
    return in_poly(*args, **kwargs)


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)





#!/usr/bin/env python


__all__ = ['lat_fmt', 'lon_fmt']


def lat_fmt(lat):
    r"""
    Set lat label string (called by Basemap.drawparallels) if LaTeX package
    clash.


    Definition
    ----------
    def lat_fmt(lat):


    Input
    -----
    latitude -90-90


    Output
    ------
    formatted string


    Examples
    --------
    >>> import matplotlib as mpl
    >>> mpl.rcParams['text.usetex'] = False
    >>> lat = [-90, -45, 0, 45, 90]
    >>> for l in lat: print(lat_fmt(l))
    90$^{\circ}$S
    45$^{\circ}$S
    0$^{\circ}$N
    45$^{\circ}$N
    90$^{\circ}$N

    >>> mpl.rcParams['text.usetex'] = True
    >>> for l in lat: print(lat_fmt(l))
    90$\/^{\circ}\/\mathrm{S}$
    45$\/^{\circ}\/\mathrm{S}$
    0$\/^{\circ}\/\mathrm{N}$
    45$\/^{\circ}\/\mathrm{N}$
    90$\/^{\circ}\/\mathrm{N}$

    History
    -------
    Written,  MC, Oct 2015
    """
    from matplotlib import rcParams
    if rcParams['text.usetex']:
        if lat < 0:
            return str(abs(int(lat))) + r'$\/^{\circ}\/\mathrm{S}$'
        else:
            return str(int(lat)) + r'$\/^{\circ}\/\mathrm{N}$'
    else:
        if lat < 0:
            return str(abs(int(lat))) + r'$^{\circ}$S'
        else:
            return str(int(lat)) + r'$^{\circ}$N'


def lon_fmt(lon):
    r"""
    Set lon label string (called by Basemap.drawmeridians) if LaTeX package
    clash.


    Definition
    ----------
    def lon_fmt(lon):


    Input
    -----
    longitude 0-360


    Output
    ------
    formatted string


    Examples
    --------
    >>> import matplotlib as mpl
    >>> mpl.rcParams['text.usetex'] = False
    >>> lon = [0, 90, 180, 270, 360]
    >>> for l in lon: print(lon_fmt(l))
    0$^{\circ}$E
    90$^{\circ}$E
    180$^{\circ}$E
    90$^{\circ}$W
    0$^{\circ}$E

    >>> mpl.rcParams['text.usetex'] = True
    >>> for l in lon: print(lon_fmt(l))
    0$\/^{\circ}\/\mathrm{E}$
    90$\/^{\circ}\/\mathrm{E}$
    180$\/^{\circ}\/\mathrm{E}$
    90$\/^{\circ}\/\mathrm{W}$
    0$\/^{\circ}\/\mathrm{E}$

    History
    -------
    Written,  MC, Oct 2015
    """
    from matplotlib import rcParams
    if lon > 180:
        ilon = lon - 360
    else:
        ilon = lon
    if rcParams['text.usetex']:
        if ilon < 0:
            return str(abs(int(ilon))) + r'$\/^{\circ}\/\mathrm{W}$'
        else:
            return str(int(ilon)) + r'$\/^{\circ}\/\mathrm{E}$'
    else:
        if ilon < 0:
            return str(abs(int(ilon))) + r'$^{\circ}$W'
        else:
            return str(int(ilon)) + r'$^{\circ}$E'


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # lon = [0, 90, 180, 270, 360]
    # for l in lon: print(lon_fmt(l))
    # # 0$^{\circ}$E
    # # 90$^{\circ}$E
    # # 180$^{\circ}$E
    # # 90$^{\circ}$W
    # # 0$^{\circ}$E

    # lat = [-90, -45, 0, 45, 90]
    # for l in lat: print(lat_fmt(l))
    # # 90$^{\circ}$S
    # # 45$^{\circ}$S
    # # 0$^{\circ}$N
    # # 45$^{\circ}$N
    # # 90$^{\circ}$N

    # import matplotlib as mpl
    # mpl.rcParams['text.usetex'] = True

    # for l in lon: print(lon_fmt(l))
    # # 0$\/^{\circ}\/\mathrm{E}$
    # # 90$\/^{\circ}\/\mathrm{E}$
    # # 180$\/^{\circ}\/\mathrm{E}$
    # # 90$\/^{\circ}\/\mathrm{W}$
    # # 0$\/^{\circ}\/\mathrm{E}$

    # for l in lat: print(lat_fmt(l))
    # # 90$\/^{\circ}\/\mathrm{S}$
    # # 45$\/^{\circ}\/\mathrm{S}$
    # # 0$\/^{\circ}\/\mathrm{N}$
    # # 45$\/^{\circ}\/\mathrm{N}$
    # # 90$\/^{\circ}\/\mathrm{N}$



#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
"""
    Functions for interacting with an open FTP connection.


    Provided functions
    ------------------
    get_binary                  - Get a binary file from an ftp connection
    get_check_binary            - Get a binary file from an ftp connection
                                  True if all bytes transfered
                                  False if local bytes differ after transfer
    get_unix_ascii              - Get a Unix/Linux ascii file from an ftp connection
    get_check_unix_ascii        - Get a Unix/Linux ascii file from an ftp connection
                                  True if all bytes transfered
                                  False if local bytes differ after transfer
    get_windows_ascii           - Get a Windows ascii file from an ftp connection
    get_check_windows_ascii     - Get a Windows ascii file from an ftp connection
                                  True if all bytes transfered
                                  False if local bytes differ after transfer
    get_names                   - Get file names in current directory on ftp connection
    get_names_dates             - Get file names and dates in current directory on ftp connection
    get_names_sizes             - Get file names and sizes in current directory on ftp connection
    get_names_times             - Wrapper for get_names_dates
    get_names_dates_sizes       - Get file names, dates and sizes in current directory on ftp connection
    get_names_times_sizes       - Wrapper for get_names_dates_sizes
    get_size                    - Get the size of a file on an ftp connection
    get_sizes                   - Get file sizes in current directory on ftp connection
    set_mtime(ftp, fname)       - Set access and modification time of local file from date information on ftp connection
    put_binary                  - Put a binary file to an ftp connection
    put_check_binary            - Put a binary file to an ftp connection
                                  True if all bytes transfered
                                  False if remote bytes differ after transfer


    Example
    -------
    # Transfer files from FTP connection if they are larger than existing local files
    import ftplib
    import os
    # open ftp connection
    ftp = ftplib.FTP("ftp.server.de")
    ftp.login("user", "password")
    fisdir = ftp.pwd()
    # go to directory
    ftp.cwd(choose/directory)
    # get all dat files with filesizes
    flsdat, flsdatsize = get_names_sizes(ftp, '*.dat')
    # if you want to get more than one file ending, e.g. .zip and .tar.gz
    fls, flssize = get_names_sizes(ftp)
    flszip     = [ i for i in fls if i.endswith('.zip') ]
    flszipsize = [ flssize[ii] for ii, i in enumerate(fls) if i.endswith('.zip') ]
    flstar     = [ i for i in fls if i.endswith('.tar.gz') ]
    flstarsize = [ flssize[ii] for ii, i in enumerate(fls) if i.endswith('.tar.gz') ]
    # local dat file names and sizes
    llsdat     = os.listdir('*.dat')
    llsdatsize = [ int(os.stat(n).st_size) for n in llsdat ]
    ldict = dict(zip(llsdat, llsdatsize))
    fdict = dict(zip(flsdat, fldatsize))
    def local_lt_ftp(name, local_dict, ftp_dict):
        if name not in local_dict:
            return True
        else:
            if local_dict[name] < ftp_dict[name]:
                return True
            else:
                return False
    for i in flsdat:
        if local_lt_ftp(i, ldict, fdict):
            if not get_check_binary(ftp, i):
                if not get_check_binary(ftp, i):
                    print('Transfer of ', i, ' failed twice.')
                    if os.path.exists(i):
                        os.remove(i)
    ftp.quit()


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python library,
    Department of Computational Hydrosystems, Helmholtz Centre for Environmental
    Research - UFZ, Leipzig, Germany.

    Copyright (c) 2014-2016 Matthias Cuntz - mc (at) macu (dot) de

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.


    History
    -------
    Written,  MC, Jun-Dec 2014
    Modified, MC, Jan 2016
"""
from .ftp import *

# Information
__author__   = "Matthias Cuntz"
__version__  = '1.1'
__revision__ = "Revision: 2468"
__date__     = 'Date: 06.01.2016'



#!/usr/bin/env python
import numpy as np
from pyjams.fsread import sread
from pyjams.fsread import fread
import csv
import os
import time as t
import shutil as sh
from scipy.optimize import fmin


def eddycorr(indir, sltdir, cfile, hfile, meteofile, outfile, novalue=-9999,
             histstep=10, attach=True, plot=False):
    '''
    Moves EddyCorr files (cfile=35_corr.csv and hfile=36_corr.csv) from sltdir
    to indir after they have been created by EddyCorr (Kolle & Rebmann, 2007).
    Time lags between wind and concentrations are plotted and user selects
    thresholds. Water lag is correlated against rH from meteofile for missing
    values. Plots and lags are saved to outfile in indir and attached to the
    meteofile for being read by EddyFlux.


    Definition
    ----------
    eddycorr(indir, sltdir, cfile, hfile, meteofile, outfile, novalue=-9999,
             histstep=10, attach=True):


    Input
    -----
    indir       str, path of the folder where results will be saved
    sltdir      str, path of the folder containing the *.slt files and EddyCorr
                files
    cfile       str, name of the carbon lag file, e.g. 35_corr.csv (EddyCorr)
    hfile       str, name of the water lag file, e.g. 36_corr.csv (EddyCorr)
    meteofile   str, path of the meteorological file for EddyFlux (in sync with
                available *.slt files, e.g. made by meteo4slt) where the lags
                will be attached
    outfile     str, name of the output file


    Optional Input
    --------------
    novalue     int, novalue in meteofile (default=-9999)
    histstep    int, histogram steps for plotting (default=10)
    attach      bool, if True, lags will be attached to meteofile.


    Output
    ------
    lags_X.csv     file containing the lags
    c_lags_X.pdf   plot of the final carbon lags
    h_lags_X.pdf   plot of the final water lags
    laglog_X_X.log log file


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python library,
    Department of Computational Hydrosystems, Helmholtz Centre for Environmental
    Research - UFZ, Leipzig, Germany.

    Copyright (c) 2014 Arndt Piayda

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.


    History
    -------
    Written,  AP, Jul 2014
    Modified, AP, Aug 2014 - major bug fix
    '''

    ############################################################################
    # move correlation files from sltdir to indir
    sh.copy(os.path.join(sltdir, cfile), indir)
    sh.copy(os.path.join(sltdir, hfile), indir)

    ############################################################################
    # reading input file
    header = sread(os.path.join(indir,cfile))[0]
    doys   = np.array(sread(os.path.join(indir,cfile), nc=1, skip=1), dtype='|S16')
    #doys   = np.array([x[5:12] for x in doys.flatten()], dtype = '|S7')
    day    = np.array([x[5:8] for x in doys.flatten()], dtype = '|S7').astype(float)
    hour   = np.array([x[8:10] for x in doys.flatten()], dtype = '|S2').astype(float)
    min    = np.array([x[10:12] for x in doys.flatten()], dtype = '|S2').astype(float)
    doysfloat   = day + (hour + min/60.)/24.
    c      = np.array(fread(os.path.join(indir,cfile), skip=1, cskip=1))
    h      = np.array(fread(os.path.join(indir,hfile), skip=1, cskip=1))
    try:
        m      = np.array(fread(meteofile, cskip=2, nc=1))
        m      = np.where(m.astype(int) == novalue, np.NaN, m)
    except TypeError:
        print('EddyCorrWarning: No meteorology file found, no fitting with rH can be done!')
        m      = False
    cout   = np.empty((np.shape(c)[0],1))
    hout   = np.empty((np.shape(h)[0],1))

    ############################################################################
    # checking for breakpoints
    print("\nBREAKPOINT CHECK:\n")
    if plot:
        import matplotlib.pyplot as plt
        import matplotlib.backends.backend_pdf as pdf
        # plot lag c
        plt.figure(1)
        plt.plot(c[:,0], c[:,2], 'bo', label='minlag (sam)')
        plt.plot(c[:,0], c[:,5], 'ro', label='maxlag (sam)')
        plt.xlabel('DOY')
        plt.ylabel('Lags [Samples]')
        plt.title('c')
        plt.axis('auto')
        plt.grid('on')
        plt.legend()
        plt.show()

    breaks = [0]
    inp = True
    while inp:
        inp = raw_input("Breakpoints? [ddd.ddd or n]: ")
        if inp.lower() == 'n':
            inp = False
        else:
            try:
                breaks += [np.min(np.where(np.abs(doysfloat-float(inp))<0.02083))]
            except (ValueError, IndexError):
                print('EddyCorrWarning: type in floats(with . not ,), integers or n, nothing else!')
            inp = True
    breaks += [np.shape(c)[0]]

    ############################################################################
    # calling the calculation function
    if np.all(m) != False:
        for i in range(len(breaks)-1):
            cout[breaks[i]:breaks[i+1],0], hout[breaks[i]:breaks[i+1],0] = calc(c[breaks[i]:breaks[i+1]],
                                                                            h[breaks[i]:breaks[i+1]],
                                                                            m[breaks[i]:breaks[i+1]],
                                                                            doys[breaks[i]:breaks[i+1]],
                                                                            histstep, indir, plot=plot)
    else:
        for i in range(len(breaks)-1):
            cout[breaks[i]:breaks[i+1],0], hout[breaks[i]:breaks[i+1],0] = calc(c[breaks[i]:breaks[i+1]],
                                                                            h[breaks[i]:breaks[i+1]],
                                                                            m,
                                                                            doys[breaks[i]:breaks[i+1]],
                                                                            histstep, indir, plot=plot)

    ################################################################################
    # writing output file
    output = csv.writer(open(os.path.join(indir,outfile), 'w'))

    for i in xrange(np.shape(cout)[0]):
        output.writerow([np.abs(cout[i][0].astype(int)), np.abs(hout[i][0].astype(int))])

    ################################################################################
    # attaching lags to meteo file
    if attach:
        meteo = np.loadtxt(meteofile, dtype='|S21', delimiter=',')
        meteo = np.concatenate((meteo, np.abs(cout.astype(int)), np.abs(hout.astype(int))), 1)
        np.savetxt(meteofile, meteo, fmt='%s', delimiter=',')

    ############################################################################
    # plot c lag output

    print("\nFINAL LAG OUTPUT:\n")
    if plot:
        fig1 = plt.figure(7)
        sub = fig1.add_subplot(111)
        sub.plot(c[:,0], cout.astype(int), 'ro')
        plt.xlabel('doy')
        plt.ylabel('lag (sam)')
        sub.set_title('c lag ouput')
        sub.axis('auto')
        sub.grid('on')

        # plot h lag output
        fig2 = plt.figure(8)
        sub = fig2.add_subplot(111)
        sub.plot(h[:,0], hout.astype(int), 'ro')
        plt.xlabel('doy')
        plt.ylabel('h lag (sam)')
        sub.set_title('h lag ouput')
        sub.axis('auto')
        sub.grid('on')
        plt.show()

        ############################################################################
        # save figures
        pp1 = pdf.PdfPages(os.path.join(indir,'c_lags%s.pdf'%(outfile[4:-4])))
        pp2 = pdf.PdfPages(os.path.join(indir,'h_lags%s.pdf' %(outfile[4:-4])))
        fig1.savefig(pp1, format='pdf')
        fig2.savefig(pp2, format='pdf')
        pp1.close()
        pp2.close()

############################################################################
# calculations
def calc(c, h, m, doys, histstep, indir, plot=False):
    '''
    part of eddycorr: plotting and fitting of hlag to rH and witing log file
    '''
    ############################################################################
    # histograms of lag distribution c
    histcmin, bin_edgescmin = np.histogram(c[:,2], bins=np.arange(np.min(c[:,2]),
                                           np.max(c[:,2])+histstep,histstep),
                                           range=None, normed=False, weights=None)
    histcmax, bin_edgescmax = np.histogram(c[:,5], bins=np.arange(np.min(c[:,5]),
                                           np.max(c[:,5])+histstep,histstep),
                                           range=None, normed=False, weights=None)

    # histograms of lag distribution h
    histhmin, bin_edgeshmin = np.histogram(h[:,2], bins=np.arange(np.min(h[:,2]),
                                           np.max(h[:,2])+histstep,histstep),
                                           range=None, normed=False, weights=None)
    histhmax, bin_edgeshmax = np.histogram(h[:,5], bins=np.arange(np.min(h[:,5]),
                                           np.max(h[:,5])+histstep,histstep),
                                           range=None, normed=False, weights=None)

    ############################################################################
    # plot lag c
    print("\nCO2 LAGS:\n")
    if plot:
        import matplotlib.pyplot as plt
        import matplotlib.backends.backend_pdf as pdf
        plt.figure(1)
        plt.plot(c[:,0], c[:,2], 'bo', label='minlag (sam)')
        plt.plot(c[:,0], c[:,5], 'ro', label='maxlag (sam)')
        plt.xlabel('DOY')
        plt.ylabel('Lags [Samples]')
        plt.title('c')
        plt.axis('auto')
        plt.grid('on')
        plt.legend()

        # plot histogram lag c
        plt.figure(2)
        plt.bar(bin_edgescmin[0:-1:1], histcmin, width=histstep/2-1, color='b', label='minlag (sam)')
        plt.bar(bin_edgescmax[0:-1:1]+histstep/2, histcmax, width=histstep/2-1, color='r', label='maxlag (sam)')
        plt.xlabel('Classes of lags [samples]')
        plt.ylabel('count')
        plt.title('c')
        plt.grid('on')
        plt.legend()
        plt.show()

    ctop = float(raw_input("Top of CO2 lag range: "))
    cbottom = float(raw_input("Bottom of CO2 lag range: "))

    ############################################################################
    # plot lag h
    print("\nH2O LAGS:\n")
    if plot:
        plt.figure(3)
        plt.plot(h[:,0], h[:,2], 'bo', label='minlag (sam)')
        plt.plot(h[:,0], h[:,5], 'ro', label='maxlag (sam)')
        plt.xlabel('DOY')
        plt.ylabel('Lags [Samples]')
        plt.title('h')
        plt.axis('auto')
        plt.grid('on')
        plt.legend()

        # plot histogram lag h
        plt.figure(4)
        plt.bar(bin_edgeshmin[0:-1:1], histhmin, width=histstep/2-1, color='b', label='minlag (sam)')
        plt.bar(bin_edgeshmax[0:-1:1]+histstep/2, histhmax, width=histstep/2-1, color='r', label='maxlag (sam)')
        plt.xlabel('Classes of lags [samples]')
        plt.ylabel('count')
        plt.title('h')
        plt.grid('on')
        plt.legend()
        plt.show()

    print("! Only maxlag (sam) can be used for H2O !")
    htop    = float(raw_input("Top of H2O lag range: "))
    hmedian = float(raw_input("Median of  H2O lag range: "))
    hbottom = float(raw_input("Bottom of H2O lag range: "))

    ############################################################################
    # preparing data for regression

    print("\nCHECKING lag-rH FIT:\n")

    hsub = h[np.where((h[:,5]<=htop)&(h[:,5]>=hbottom))[0],5]
    msub = m[np.where((h[:,5]<=htop)&(h[:,5]>=hbottom))[0],0]
    if plot:
        # plot max lag vs. rH
        fig3 = plt.figure(6)
        sub = fig3.add_subplot(111)
        sub.plot(msub, hsub, 'ro', label='measurements')
        plt.xlabel('rH [%]')
        plt.ylabel('maxlag (sam)')
        sub.set_title('h maxlag (sam) - rH correlation')
        sub.axis('auto')
        sub.grid('on')
        plt.legend()
        plt.show()

    pbt = 't'
    while pbt=='t':
        mmax = float(raw_input("Top of rel. hum: "))
        hmin = float(raw_input("Maximum of  H2O lag: "))
        valid=(msub<mmax) & (hsub<hmin)
        p_guess1 = float(raw_input("Initial guess - Lag offset: "))*-1.
        p_guess2 = float(raw_input("Initial guess - rH multiplier: "))

        p, ff = fit(-hsub[valid], msub[valid], func=f, p_guess=[p_guess1,p_guess2],plot=plot)
        print('Offset=%f'%(p[0]*-1.), 'Multiplier=%f'%(p[1]))
        pbt = raw_input("(g)ood fit - proceed, (b)ad fit - proceed, (t)ry again: ")
        if pbt not in ['g', 'b', 't']:
            pbt = 't'

    ############################################################################
    # lags for c for output
    cout = np.where((c[:,2]<=ctop) & (c[:,2]>=cbottom), c[:,2],
                    np.where((c[:,5]<=ctop) & (c[:,5]>=cbottom), c[:,5],
                             (ctop+cbottom)/2))

    # lags for h for output
    if pbt=="g":
        hout = np.where((h[:,5]>htop)^(h[:,5]<hbottom), -f_inv(m[:,0], p), h[:,5])
        hout = np.where((hout>htop)^(hout<hbottom), hmedian, hout)
        hout = np.where(np.isnan(hout), hmedian, hout)
    else:
        hout = np.where((h[:,5]>htop)^(h[:,5]<hbottom), hmedian, h[:,5])
        hout = np.where(np.isnan(hout), hmedian, hout)

    ################################################################################
    # writing log file
    log = open(os.path.join(indir,'lag_%i_%02i_%02i_%02i_%02i_%02i.log' %(
                                                    t.localtime()[0], t.localtime()[1],
                                                    t.localtime()[2], t.localtime()[3],
                                                    t.localtime()[4], t.localtime()[5],))
                                                    , 'w')
    log.write('From doy %s %s:%s to doy %s %s:%s: \n' %(doys.flatten()[0][5:8], doys.flatten()[0][8:10], doys.flatten()[0][10:12], doys.flatten()[-1][5:8], doys.flatten()[-1][8:10], doys.flatten()[-1][10:12]))
    log.write('Top of CO2 lag range: %i\n' %(int(ctop)))
    log.write('Bottom of CO2 lag range: %i\n' %(int(cbottom)))
    log.write('Top of H2O lag range: %i\n' %(int(htop)))
    log.write('Bottom of H2O lag range: %i\n' %(int(hbottom)))
    log.write('Median of  H2O lag range: %i\n' %(int(hmedian)))
    if pbt=="g":
        log.write('Moisture dependency:\n')
        log.write('Top of rH used for fitting: %f\n'%mmax)
        log.write('Maximum of  H2O lag used for fitting: %f\n'%hmin)
        log.write('Initial guess - Lag offset: %f\n'%(p_guess1*-1.))
        log.write('Initial guess - rH multiplier: %f\n'%p_guess2)
        log.write('Moisture dependency: -maxlag_h[sam] = %f + exp(rH/%f)'%(p[0], p[1]))
    else:
        log.write('Median used instead of moisture dependency.')

    return cout, hout

############################################################################
# fitting
def fit(x,y,func,p_guess,plot=False):
    '''
    eddycorr: fitting procedure
    '''
    xfit=np.ma.array(x, mask=np.isnan(x))
    yfit=np.ma.array(y, mask=np.isnan(y))

    if plot:
        import matplotlib.pyplot as plt
        import matplotlib.backends.backend_pdf as pdf
        x_mod = np.arange(np.ma.min(xfit),np.ma.max(xfit),(np.ma.max(xfit)-np.ma.min(xfit))/50.)
        plt.figure('pre-fit')
        plt.plot(xfit, yfit, 'bo')
        plt.plot(x_mod, func(x_mod,p_guess), 'k-')
        plt.show()

    # fit
    p_opt = fmin(absdif, p_guess, args=(xfit, yfit, func), disp=0)

    if plot:
        plt.figure('post-fit')
        plt.plot(xfit, yfit, 'bo')
        plt.plot(x_mod, func(x_mod,p_opt), 'k-')
        plt.show()

    return p_opt, func

############################################################################
# h lag model
def f(x, p):
    '''
    eddycorr: model for relating relative humidity and -maxlag_h
    '''
    return p[1]*np.log(x-p[0])

def f_inv(x, p):
    '''
    eddycorr: model for relating relative humidity and -maxlag_h
    '''
    return p[0]+np.exp(x/p[1])

def absdif(p,x,y,func):
    '''
    eddycorr: objective function
    '''
    return np.sum(np.abs(y-func(x, p)))

if __name__ == '__main__':
    import doctest
    doctest.testmod()



#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
import numpy as np
from math import factorial
import scipy.signal as sps


__all__ = ['savitzky_golay', 'sg', 'savitzky_golay2d', 'sg2d']


def savitzky_golay(y, window, order, deriv=0, rate=1):
    """
    Smooth (and optionally differentiate) data with a Savitzky-Golay filter.
    The Savitzky-Golay filter removes high frequency noise from data.
    It has the advantage of preserving the original shape and
    features of the signal better than other types of filtering
    approaches, such as moving averages techniques.


    Definition
    ----------
    def savitzky_golay(y, window, order, deriv=0, rate=1):


    Input
    -----
    y         array_like, shape (N,)
              the values of the time history of the signal.
    window    int
              the length of the window. Must be an odd integer number.
    order     int
              the order of the polynomial used in the filtering.
              Must be less then `window` - 1.


    Optional Input
    --------------
    deriv : int
        the order of the derivative to compute
        (default = 0 means only smoothing)
    rate : int
        ??? output will be multiplied by rate**deriv

    Output
    ------
    ndarray : shape(N,)
        the smoothed signal (or it's n-th derivative).


    Notes
    -----
    The Savitzky-Golay is a type of low-pass filter, particularly
    suited for smoothing noisy data. The main idea behind this
    approach is to make for each point a least-square fit with a
    polynomial of high order over a odd-sized window centered at
    the point.


    References
    ----------
    [1] A. Savitzky, M. J. E. Golay
        Smoothing and Differentiation of Data by Simplified Least Squares
        Procedures. Analytical Chemistry, 1964, 36 (8), pp 1627-1639.
    [2] W.H. Press, S.A. Teukolsky, W.T. Vetterling, B.P. Flannery
        Numerical Recipes 3rd Edition: The Art of Scientific Computing
        Cambridge University Press ISBN-13: 9780521880688


    Examples
    --------
    >>> import numpy as np
    >>> np.random.seed(1)
    >>> t = np.linspace(-4, 4, 10)
    >>> y = np.exp( -t**2 ) + np.random.normal(0, 0.05, t.shape)
    >>> from autostring import astr
    >>> print(astr(savitzky_golay(y, window=3, order=1),3,pp=True))
    [' 0.007' ' 0.010' ' 0.022' ' 0.320' ' 0.562' ' 0.609' ' 0.310' ' 0.080' '-0.009' ' 0.007']

    >>> print(astr(savitzky_golay(y, 3, 1, deriv=1),3,pp=True))
    [' 0.000' '-0.050' ' 0.073' ' 0.442' ' 0.295' '-0.304' '-0.368' '-0.120' ' 0.009' ' 0.000']

    >>> print(astr(savitzky_golay(y, 3, 1, deriv=1, rate=10),3,pp=True))
    [' 0.000' '-0.502' ' 0.729' ' 4.416' ' 2.952' '-3.039' '-3.683' '-1.201' ' 0.092' ' 0.000']


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python
    library, Department of Computational Hydrosystems, Helmholtz Centre for
    Environmental Research - UFZ, Leipzig, Germany.

    Copyright (c) 2012-2021 Matthias Cuntz - mc (at) macu (dot) de

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.


    History
    -------
    Written,  Matthias Cuntz, Oct 2012
                  - from SciPy cookbook:
                    http://www.scipy.org/Cookbook/SavitzkyGolay
    Modified, Matthias Cuntz, Feb 2013 - ported to Python 3
              Matthias Cuntz, Apr 2014 - assert
              Matthias Cuntz, Sep 2021 - code refactoring
    """
    #
    # Check input
    try:
        window = np.abs(int(window))
        order  = np.abs(int(order))
    except ValueError as msg:
        raise ValueError("window and order have to be of type int.")
    assert ((window % 2 == 1) and (window >= 1)), "window size must be a positive odd number."
    assert window >= (order + 2), "window is too small for polynomial order."
    order_range = list(range(order+1))
    half_window = (window-1) // 2
    #
    # precompute coefficients (m)
    b = np.mat([[k**i for i in order_range]
                for k in range(-half_window, half_window+1)])
    m = np.linalg.pinv(b).A[deriv] * rate**deriv * factorial(deriv)
    #
    # pad the signal at the extremes with values taken from the signal itself
    firstvals = y[0]  - np.abs(y[1:half_window+1][::-1]   - y[0])
    lastvals  = y[-1] + np.abs(y[-half_window-1:-1][::-1] - y[-1])
    y = np.concatenate((firstvals, y, lastvals))
    #
    # Filter with pre-computed coefficients m
    return np.convolve(m[::-1], y, mode='valid')


def sg(y, window, order, deriv=0, rate=1):
    """
    Wrapper function for savitzky_golay
    def sg(y, window, order, deriv=0, rate=1):


   Examples
   --------
   >>> import numpy as np
   >>> np.random.seed(1)
   >>> t = np.linspace(-4, 4, 10)
   >>> y = np.exp( -t**2 ) + np.random.normal(0, 0.05, t.shape)
   >>> from autostring import astr
   >>> print(astr(sg(y, window=3, order=1),3,pp=True))
   [' 0.007' ' 0.010' ' 0.022' ' 0.320' ' 0.562' ' 0.609' ' 0.310' ' 0.080' '-0.009' ' 0.007']

   >>> print(astr(sg(y, 3, 1, deriv=1),3,pp=True))
   [' 0.000' '-0.050' ' 0.073' ' 0.442' ' 0.295' '-0.304' '-0.368' '-0.120' ' 0.009' ' 0.000']

   >>> print(astr(sg(y, 3, 1, deriv=1, rate=10),3,pp=True))
   [' 0.000' '-0.502' ' 0.729' ' 4.416' ' 2.952' '-3.039' '-3.683' '-1.201' ' 0.092' ' 0.000']
    """
    return savitzky_golay(y, window, order, deriv=deriv, rate=rate)


def savitzky_golay2d(z, window, order, deriv=None):
    """
    Smooth (and optionally differentiate) 2D data with a Savitzky-Golay filter.
    The Savitzky-Golay filter removes high frequency noise from data.
    It has the advantage of preserving the original shape and
    features of the signal better than other types of filtering
    approaches, such as moving averages techniques.


    Definition
    ----------
    def savitzky_golay2d(z, window, order, deriv=None)


    Input
    -----
    z         array_like, shape (N,M)
              the value array.
    window    int
              the length of the window. Must be an odd integer number.
    order     int
              the order of the polynomial used in the filtering.
              Must be less then `window` - 1.


    Optional Input
    --------------
    deriv     None=0, 'both=1, 'col'=2, 'row'=3
              None or 0: normal smoothing
              'both' or 1: returns the gradient (first derivatives)
              'col' or 2, or 'row' or 3: indicates the direction of
                                         the derivative

    Output
    ------
    ndarray : shape(N,M) if deriv=0, 2 or 3;
              tuple(shape(N,M),shape(N,M)) if deriv=1
        the smoothed signal, or it's 1st derivative in one direction or both.


    Notes
    -----
    The Savitzky-Golay is a type of low-pass filter, particularly
    suited for smoothing noisy data. The main idea behind this
    approach is to make for each point a least-square fit with a
    polynomial of high order over a odd-sized window centered at
    the point.

    Savitsky-Golay filters can also be used to smooth two dimensional
    data affected by noise. The algorithm is exactly the same as for
    the one dimensional case, only the math is a bit more tricky. The
    basic algorithm is as follow:
    1. for each point of the two dimensional matrix extract a sub-matrix,
       centered at that point and with a size equal to an odd number
       "window_size".
    2. for this sub-matrix compute a least-square fit of a polynomial surface,
       defined as
         p(x,y) = a0 + a1*x + a2*y + a3*x2 + a4*y2 + a5*x*y + ... .
       Note that x and y are equal to zero at the central point.
    3. replace the initial central point with the value computed with the fit.


    References
    ----------
    [1] A. Savitzky, M. J. E. Golay
        Smoothing and Differentiation of Data by Simplified Least Squares
        Procedures. Analytical Chemistry, 1964, 36 (8), pp 1627-1639.
    [2] W.H. Press, S.A. Teukolsky, W.T. Vetterling, B.P. Flannery
        Numerical Recipes 3rd Edition: The Art of Scientific Computing
        Cambridge University Press ISBN-13: 9780521880688


    Examples
    --------
    >>> import numpy as np
    >>> np.random.seed(1)
    >>> x = np.linspace(-3,3,5)
    >>> y = np.linspace(-3,3,3)
    >>> X, Y = np.meshgrid(x,y)
    >>> Z = np.exp( -(X**2+Y**2))
    >>> Zn = Z + np.random.normal(0, 0.2, Z.shape)
    >>> from autostring import astr
    >>> print(astr(savitzky_golay2d(Zn, window=3, order=2),3,pp=True))
    [[' 0.080' ' 0.052' '-0.106' '-0.165' ' 0.161']
     ['-0.204' ' 0.170' ' 0.664' ' 0.181' ' 0.048']
     [' 0.111' '-0.230' ' 0.094' '-0.170' ' 0.240']]

    >>> s2 = savitzky_golay2d( Zn, 3, 2, deriv=1)
    >>> print(astr(s2[0],3,pp=True))
    [[' 0.156' ' 0.046' '-0.046' ' 0.065' ' 0.257']
     [' 0.305' ' 0.087' '-0.007' '-0.055' ' 0.230']
     [' 0.305' ' 0.126' '-0.039' '-0.158' ' 0.120']]

    >>> s3 = savitzky_golay2d( Zn, 3, 2, deriv='both')
    >>> if np.any(s2[0]!=s3[0]): print('Error 01')
    >>> if np.any(s2[1]!=s3[1]): print('Error 02')

    >>> s2 = savitzky_golay2d( Zn, 3, 2, deriv=2)
    >>> print(astr(s2,3,pp=True))
    [['-0.069' ' 0.510' ' 0.638' ' 0.446' ' 0.128']
     ['-0.102' '-0.047' '-0.018' ' 0.039' ' 0.027']
     [' 0.556' ' 0.251' ' 0.000' ' 0.092' ' 0.111']]

    >>> s3 = savitzky_golay2d( Zn, 3, 2, deriv='col')
    >>> if np.any(s2!=s3): print('Error 03')


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License.

    Copyright (c) 2012-2021 Matthias Cuntz - mc (at) macu (dot) de

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.


    History
    -------
    Written,  Matthias Cuntz, Oct 2012
                  - from SciPy cookbook:
                    http://www.scipy.org/Cookbook/SavitzkyGolay
    Modified, Matthias Cuntz, Nov 2012
                  - replaced fftconvolve by convolve because of crash on Mac
              Matthias Cuntz, Feb 2013 - ported to Python 3
              Matthias Cuntz, Apr 2014 - assert
              Matthias Cuntz, Sep 2021 - code refactoring
    """
    #
    # number of terms in the polynomial expression
    n_terms = ( order + 1 ) * ( order + 2)  / 2.0
    assert window % 2 == 1, 'window must be odd.'
    assert window**2 >= n_terms, 'order is too high for the window size.'
    half_size = window // 2
    #
    # exponents of the polynomial.
    # p(x,y) = a0 + a1*x + a2*y + a3*x^2 + a4*y^2 + a5*x*y + ...
    # this line gives a list of two item tuple. Each tuple contains
    # the exponents of the k-th term. First element of tuple is for x
    # second element for y.
    # Ex. exps = [(0,0), (1,0), (0,1), (2,0), (1,1), (0,2), ...]
    exps = [ (k-n, n) for k in range(order+1) for n in range(k+1) ]
    #
    # coordinates of points
    ind = np.arange(-half_size, half_size+1, dtype=np.float64)
    dx = np.repeat( ind, window )
    dy = np.tile( ind, [window, 1]).reshape(window**2, )
    #
    # build matrix of system of equation
    A = np.empty( (window**2, len(exps)) )
    for i, exp in enumerate( exps ):
        A[:, i] = (dx**exp[0]) * (dy**exp[1])
    #
    # pad input array with appropriate values at the four borders
    new_shape = z.shape[0] + 2*half_size, z.shape[1] + 2*half_size
    Z = np.zeros( (new_shape) )
    # top band
    band = z[0, :]
    Z[:half_size, half_size:-half_size] = band - np.abs(
        np.flipud( z[1:half_size+1, :] ) - band )
    # bottom band
    band = z[-1, :]
    Z[-half_size:, half_size:-half_size] = band + np.abs( np.flipud(
        z[-half_size-1:-1, :] ) - band )
    # left band
    band = np.tile( z[:, 0].reshape(-1, 1), [1, half_size])
    Z[half_size:-half_size, :half_size] = band - np.abs(
        np.fliplr( z[:, 1:half_size+1] ) - band )
    # right band
    band = np.tile( z[:, -1].reshape(-1, 1), [1, half_size] )
    Z[half_size:-half_size, -half_size:] = band + np.abs(
        np.fliplr( z[:, -half_size-1:-1] ) - band )
    # central band
    Z[half_size:-half_size, half_size:-half_size] = z
    #
    # pad input array with appropriate values at the four corners
    # top left corner
    band = z[0, 0]
    Z[:half_size, :half_size] = band - np.abs(
        np.flipud(np.fliplr(z[1:half_size+1, 1:half_size+1]) ) - band )
    # bottom right corner
    band = z[-1, -1]
    Z[-half_size:, -half_size:] = band + np.abs(
        np.flipud(np.fliplr(z[-half_size-1:-1, -half_size-1:-1]) ) - band )
    # top right corner
    band = Z[half_size, -half_size:]
    Z[:half_size, -half_size:] = band - np.abs(
        np.flipud(Z[half_size+1:2*half_size+1, -half_size:]) - band )
    # bottom left corner
    band = Z[-half_size:, half_size].reshape(-1, 1)
    Z[-half_size:, :half_size] = band - np.abs(
        np.fliplr(Z[-half_size:, half_size+1:2*half_size+1]) - band )
    #
    # solve system and convolve
    # crashed on example at Mac OSX 10.7.5 with Python 2.7.1
    if (deriv is None) | (deriv == 0):
        m = np.linalg.pinv(A)[0].reshape((window, -1))
        # return sps.fftconvolve(Z, m, mode='valid')
        return sps.convolve(Z, m, mode='valid')
    elif (deriv == 'both') | (deriv == 1):
        c = np.linalg.pinv(A)[1].reshape((window, -1))
        r = np.linalg.pinv(A)[2].reshape((window, -1))
        # return (sps.fftconvolve(Z, -r, mode='valid'),
        #         sps.fftconvolve(Z, -c, mode='valid'))
        return (sps.convolve(Z, -r, mode='valid'),
                sps.convolve(Z, -c, mode='valid'))
    elif (deriv == 'col') | (deriv == 2):
        c = np.linalg.pinv(A)[1].reshape((window, -1))
        # return sps.fftconvolve(Z, -c, mode='valid')
        return sps.convolve(Z, -c, mode='valid')
    elif (deriv == 'row') | (deriv == 3):
        r = np.linalg.pinv(A)[2].reshape((window, -1))
        # return sps.fftconvolve(Z, -r, mode='valid')
        return sps.convolve(Z, -r, mode='valid')


def sg2d(*args, **kwargs):
    """
    Wrapper function for savitzky_golay2d
    def savitzky_golay2d(z, window, order, deriv=None):


    Examples
    --------
    >>> import numpy as np
    >>> np.random.seed(1)
    >>> x = np.linspace(-3,3,5)
    >>> y = np.linspace(-3,3,3)
    >>> X, Y = np.meshgrid(x,y)
    >>> Z = np.exp( -(X**2+Y**2))
    >>> Zn = Z + np.random.normal(0, 0.2, Z.shape)
    >>> from autostring import astr
    >>> print(astr(sg2d(Zn, window=3, order=2),3,pp=True))
    [[' 0.080' ' 0.052' '-0.106' '-0.165' ' 0.161']
     ['-0.204' ' 0.170' ' 0.664' ' 0.181' ' 0.048']
     [' 0.111' '-0.230' ' 0.094' '-0.170' ' 0.240']]

    >>> s2 = sg2d( Zn, 3, 2, deriv=1)
    >>> print(astr(s2[0],3,pp=True))
    [[' 0.156' ' 0.046' '-0.046' ' 0.065' ' 0.257']
     [' 0.305' ' 0.087' '-0.007' '-0.055' ' 0.230']
     [' 0.305' ' 0.126' '-0.039' '-0.158' ' 0.120']]

    >>> s3 = sg2d( Zn, 3, 2, deriv='both')
    >>> if np.any(s2[0]!=s3[0]): print('Error 01')
    >>> if np.any(s2[1]!=s3[1]): print('Error 02')

    >>> s2 = sg2d( Zn, 3, 2, deriv=2)
    >>> print(astr(s2,3,pp=True))
    [['-0.069' ' 0.510' ' 0.638' ' 0.446' ' 0.128']
     ['-0.102' '-0.047' '-0.018' ' 0.039' ' 0.027']
     [' 0.556' ' 0.251' ' 0.000' ' 0.092' ' 0.111']]

    >>> s3 = sg2d( Zn, 3, 2, deriv='col')
    >>> if np.any(s2!=s3): print('Error 03')
    """
    return savitzky_golay2d(*args, **kwargs)


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # import matplotlib.pyplot as plt
    # #
    # # 1D
    # t = np.linspace(-4, 4, 500)
    # y = np.exp( -t**2 ) + np.random.normal(0, 0.05, t.shape)
    # ysg = sg(y, window=31, order=4)

    # plt.figure(1)
    # plt.plot(t, y, label='Noisy signal')
    # plt.plot(t, np.exp(-t**2), 'k', lw=1.5, label='Original signal')
    # plt.plot(t, ysg, 'r', label='Filtered signal')
    # plt.legend()

    # #
    # # 2D
    # x = np.linspace(-3,3,100)
    # y = np.linspace(-3,3,200)
    # X, Y = np.meshgrid(x,y)
    # Z = np.exp( -(X**2+Y**2))
    # Zn = Z + np.random.normal(0, 0.2, Z.shape)
    # Zf = sg2d(Zn, window=29, order=4)

    # plt.matshow(Z)
    # plt.matshow(Zn)
    # plt.matshow(Zf)

    # plt.show()

    # np.random.seed(1)
    # t = np.linspace(-4, 4, 10)
    # y = np.exp( -t**2 ) + np.random.normal(0, 0.05, t.shape)
    # print savitzky_golay(y, window=3, order=1)
    # #[ 0.0067223   0.01048351  0.02186601  0.32004949  0.56168932  0.60865232
    # #  0.31034614  0.08045836 -0.00911575  0.00652022]

    # print savitzky_golay(y, 3, 1, deriv=1)
    # #[  0.00000000e+00  -5.02294966e-02   7.29450608e-02   4.41633401e-01
    # #   2.95156495e-01  -3.03885643e-01  -3.68285622e-01  -1.20119683e-01
    # #   9.21248188e-03  -1.73472348e-18]

    # print savitzky_golay(y, 3, 1, deriv=1, rate=10)
    # #[ -2.77555756e-17  -5.02294966e-01   7.29450608e-01   4.41633401e+00
    # #  2.95156495e+00  -3.03885643e+00  -3.68285622e+00  -1.20119683e+00
    # #  9.21248188e-02  -1.38777878e-17]

    # np.random.seed(1)
    # x = np.linspace(-3,3,5)
    # y = np.linspace(-3,3,3)
    # X, Y = np.meshgrid(x,y)
    # Z = np.exp( -(X**2+Y**2))
    # Zn = Z + np.random.normal(0, 0.2, Z.shape)
    # print sg2d( Zn, window=3, order=2)
    # #[[ 0.07980947  0.05211804 -0.10551094 -0.16506245  0.16105317]
    # # [-0.20426342  0.17009221  0.66354671  0.18103661  0.04756387]
    # # [ 0.11059771 -0.22965185  0.09383012 -0.17015538  0.23954209]]

    # s2 = sg2d( Zn, 3, 2, deriv=1)
    # print s2[0]
    # #[[ 0.15577951  0.04649446 -0.04612122  0.06501884  0.25715001]
    # # [ 0.30484864  0.08679688 -0.00702995 -0.05463381  0.23040468]
    # # [ 0.30484864  0.12564873 -0.03918197 -0.15848258  0.12036621]]
    # print s2[1]
    # #[[-0.06945119  0.5099898   0.63791906  0.44568577  0.12792925]
    # # [-0.10196687 -0.04682891 -0.01845718  0.03876769  0.02683618]
    # # [ 0.55571728  0.25086864  0.          0.09216819  0.11135048]]

    # s3 = sg2d( Zn, 3, 2, deriv='both')
    # if np.any(s2[0]!=s3[0]): print 'Error 01'
    # if np.any(s2[1]!=s3[1]): print 'Error 02'

    # s2 = sg2d( Zn, 3, 2, deriv=2)
    # print s2
    # #[[-0.06945119  0.5099898   0.63791906  0.44568577  0.12792925]
    # # [-0.10196687 -0.04682891 -0.01845718  0.03876769  0.02683618]
    # # [ 0.55571728  0.25086864  0.          0.09216819  0.11135048]]

    # s3 = sg2d( Zn, 3, 2, deriv='col')
    # if np.any(s2!=s3): print 'Error 03'
